#ifndef json__marshaling__mapping_hxx_
#define json__marshaling__mapping_hxx_

#include <code/json/optional.hxx>
#include <code/json/pointer.hxx>
#include <code/json/variant.hxx>

#include <code/json/marshaling/marshaling-traits.hxx>
#include <code/json/marshaling/traits.hxx>

#include <functional>
#include <variant>

namespace code::json::marshaling {

template< typename T >
class member_mapping {
public:
  using getter_type = std::function< variant(T const&, marshaling_context*) >;
  using setter_type = std::function< void(T&, variant const&, marshaling_context*) >;

  member_mapping(std::string key,
                 bool optional,
                 getter_type getter,
                 setter_type setter)
    : key_{ std::move(key) },
      optional_{ optional },
      getter_{ std::move(getter) },
      setter_{ std::move(setter) }
  {}

  member_mapping(pointer p,
                 bool optional,
                 getter_type getter,
                 setter_type setter)
    : key_{ std::move(p) },
      optional_{ optional },
      getter_{ std::move(getter) },
      setter_{ std::move(setter) }
  {}

  template< typename U >
  member_mapping(member_mapping< U > const& other)
    : key_{other.key_},
      optional_{other.optional_},
      getter_{other.getter_},
      setter_{other.setter_}
  {}

  std::variant<std::string, pointer> const&
  key() const
  {
    return key_;
  }

  bool
  optional() const
  {
    return optional_;
  }

  variant
  get(T const& instance, marshaling_context* context) const
  {
    return getter_(instance, context);
  }

  void
  set(T& instance, variant const& value, marshaling_context* context) const
  {
    setter_(instance, value, context);
  }

private:
  template< typename U >
  friend class member_mapping;

  std::variant<std::string, pointer> key_;
  bool optional_;
  getter_type getter_;
  setter_type setter_;
};

template< typename T, typename M >
member_mapping< T >
member(std::string key, M T::*member)
{
  return member_mapping< T >{
    std::move(key),
    is_optional_v<M>,
    [member](T const& instance, marshaling_context* context) -> json::variant
    {
      return marshaling_traits< M >::marshal(instance.*member, context);
    },
    [member](T& instance, json::variant const& v, marshaling_context* context)
    {
      instance.*member = marshaling_traits< M >::unmarshal(v, context);
    }
  };
}

template< typename T, typename M >
member_mapping< T >
member(pointer p, M T::*member)
{
  return member_mapping< T >{
    std::move(p),
    is_optional_v<M>,
    [member](T const& instance, marshaling_context* context) -> json::variant
    {
      return marshaling_traits< M >::marshal(instance.*member, context);
    },
    [member](T& instance, json::variant const& v, marshaling_context* context)
    {
      instance.*member = marshaling_traits< M >::unmarshal(v, context);
    }
  };
}

template< typename T, typename Accessor >
member_mapping< T >
accessor(std::string key, Accessor access)
{
  return member_mapping< T >{
    std::move(key),
    is_optional_v<std::decay_t<decltype(access(std::declval<T>()))>>,
    [access](T const& instance, marshaling_context* context) -> json::variant
    {
      using M = std::decay_t<decltype(access(instance))>;
      return marshaling_traits< M >::marshal(access(instance), context);
    },
    [access](T& instance, json::variant const& v, marshaling_context* context)
    {
      using M = std::decay_t<decltype(access(instance))>;
      access(instance) = marshaling_traits< M >::unmarshal(v, context);
    }
  };
}

template< typename T, typename Accessor >
member_mapping< T >
accessor(pointer p, Accessor access)
{
  return member_mapping< T >{
    std::move(p),
    is_optional_v<std::decay_t<decltype(access(std::declval<T>()))>>,
    [access](T const& instance, marshaling_context* context) -> json::variant
    {
      using M = std::decay_t<decltype(access(instance))>;
      return marshaling_traits< M >::marshal(access(instance), context);
    },
    [access](T& instance, json::variant const& v, marshaling_context* context)
    {
      using M = std::decay_t<decltype(access(instance))>;
      access(instance) = marshaling_traits< M >::unmarshal(v, context);
    }
  };
}

template< typename T, typename V >
member_mapping< T >
static_value(std::string key, V value)
{
  return member_mapping< T >{
    std::move(key),
    true,
    [value](T const& instance, marshaling_context* context) -> json::variant {
      return marshaling_traits< V >::marshal(value, context);
    },
    [](T&, json::variant const&, marshaling_context*) {
      // No-op.
    }
  };
}

template< typename T, typename V >
member_mapping< T >
static_value(pointer p, V value)
{
  return member_mapping< T >{
    std::move(p),
    true,
    [value](T const& instance, marshaling_context* context) -> json::variant {
      return marshaling_traits< V >::marshal(value, context);
    },
    [](T&, json::variant const&, marshaling_context*) {
      // No-op.
    }
  };
}

template< typename T >
class mapping {
public:
  using container_type = std::vector< member_mapping< T > >;
  using const_iterator = typename container_type::const_iterator;

  mapping(container_type mappings)
    : mappings_{std::move(mappings)}
  {}

  mapping(std::initializer_list< member_mapping< T > > init)
  {
    for (auto const& j : init)
      mappings_.emplace_back(j);
  }

  const_iterator
  begin() const
  {
    return mappings_.begin();
  }

  const_iterator
  end() const
  {
    return mappings_.end();
  }

private:
  container_type mappings_;

};

template<typename T, typename... Bases>
mapping<T>
map(std::initializer_list<member_mapping<T>> init)
{
  typename mapping<T>::container_type m{init};

  auto inherit = [&](auto&& base)
  {
    for (auto const& j : base) {
      m.emplace_back(j);
    }
  };

  ((inherit(Bases::json())), ...);

  return m;
}

} // namespace code::json::marshaling

#endif
