#ifndef code__json__marshaling__marshaling_traits_hxx_
#define code__json__marshaling__marshaling_traits_hxx_

#include <code/json/optional.hxx>
#include <code/json/pointer.hxx>
#include <code/json/variant.hxx>

#include <code/json/marshaling/marshaling-context.hxx>

#include <chrono>
#include <ctime>
#include <deque>
#include <iomanip>
#include <iostream> // TODO: Remove.
#include <list>
#include <locale>
#include <map>
#include <set>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <variant>
#include <vector>

#include <time.h>

namespace code::json::marshaling {

template< typename T >
struct marshaling_traits {
  using model_type = T;

  static variant
  marshal(model_type const& model, marshaling_context* context)
  {
    variant v{std::map<std::string, variant>{}};

    // TODO: Handle mappings with pointers.
    for (auto const& mapping : model_type::json()) {
      if (std::holds_alternative<pointer>(mapping.key())) {
        auto ptr = std::get<pointer>(mapping.key());
        ptr.write(v, mapping.get(model, context));
      }
      else {
        v.set(std::get<std::string>(mapping.key()), mapping.get(model, context));
      }
    }

    return v;
  }

  static model_type
  unmarshal(variant const& value, marshaling_context* context)
  {
    if constexpr (std::is_default_constructible_v< model_type >) {
      // TODO: Change exception type.
      if (!value.is_object())
        throw std::runtime_error{ "cannot unmarshal non-object value" };

      model_type model;
      for (auto const& mapping : model_type::json()) {
        if (std::holds_alternative<pointer>(mapping.key())) {
          auto ptr = std::get<pointer>(mapping.key());
          auto v = ptr.read(value);

          if (!v && !mapping.optional())
            // FIXME: Add ptr to exception.
            throw std::runtime_error{ "missing field '" + to_string(ptr) + "'" };
          else
            mapping.set(model, *v, context);
          continue;
        }

        auto const& key = std::get<std::string>(mapping.key());

        if (!value.contains(key)) {
          if (!mapping.optional())
            // TODO: Change exception type.
            throw std::runtime_error{ "missing field '" + key + "'" };

          continue; // Try next key.
        }

        mapping.set(model, value.get(key), context);
      }

      return model;
    }
    else {
      throw std::runtime_error{ "this shouldn't compile" };
    }
  }
};

template< typename T >
struct marshaling_traits< optional< T > > {
  using model_type = optional< T >;

  static variant
  marshal(optional< T > const& model, marshaling_context* context)
  {
    if (model)
      return marshaling_traits< T >::marshal(*model, context);

    return {};
  }

  static optional< T >
  unmarshal(variant const& value, marshaling_context* context)
  {
    if (value.is_undefined())
      return {};

    return marshaling_traits< T >::unmarshal(value, context);
  }
};

template< typename... T >
struct marshaling_traits< std::variant< T... > > {
  using model_type = std::variant< T... >;

  static variant
  marshal(std::variant< T... > const& model, marshaling_context* context)
  {
    variant result;

    auto visitor = [&result, context](const auto& obj)
    {
      using type = std::decay_t<decltype(obj)>;
      result = marshaling_traits<type>::marshal(obj, context);
      result.set("$type", std::decay_t<decltype(obj)>::type_identifier);
    };

    std::visit(visitor, model);

    return result;
  }

  template<typename Current, typename... Next>
  struct unmarshaler
  {
    static
    std::variant<T...>
    unmarshal(variant const& value, marshaling_context* context)
    {
      if (!value.contains("$type"))
        throw std::invalid_argument{"variant missing '$type' identifier"};

      if (value.get("$type").get_string() == Current::type_identifier)
        return marshaling_traits< Current >::unmarshal(value, context);

      if constexpr (sizeof...(Next) > 0)
        return unmarshaler<Next...>::unmarshal(value, context);

      throw std::invalid_argument{"couldn't unmarshal value"};
    }
  };

  static std::variant< T... >
  unmarshal(variant const& value, marshaling_context* context)
  {
    return unmarshaler<T...>::unmarshal(value, context);
  }
};

template<>
struct marshaling_traits< bool > {
  using model_type = bool;

  static variant
  marshal(bool model, marshaling_context*)
  {
    return model;
  }

  static bool
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_boolean())
      throw std::runtime_error{ "not a boolean" };

    return value.get_boolean();
  }
};

template<>
struct marshaling_traits< short int > {
  using model_type = short int;

  static variant
  marshal(short int model, marshaling_context*)
  {
    return model;
  }

  static short int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< short int >();
  }
};

template<>
struct marshaling_traits< int > {
  using model_type = int;

  static variant
  marshal(int model, marshaling_context*)
  {
    return model;
  }

  static int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< int >();
  }
};

template<>
struct marshaling_traits< long int > {
  using model_type = long int;

  static variant
  marshal(long int model, marshaling_context*)
  {
    return model;
  }

  static long int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< long int >();
  }
};

template<>
struct marshaling_traits< long long int > {
  using model_type = long long int;

  static variant
  marshal(long long int model, marshaling_context*)
  {
    return model;
  }

  static long long int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< long long int >();
  }
};

template<>
struct marshaling_traits< unsigned short int > {
  using model_type = unsigned short int;

  static variant
  marshal(unsigned short int model, marshaling_context*)
  {
    return model;
  }

  static unsigned short int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< unsigned short int >();
  }
};

template<>
struct marshaling_traits< unsigned int > {
  using model_type = unsigned int;

  static variant
  marshal(unsigned int model, marshaling_context*)
  {
    return model;
  }

  static unsigned int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< unsigned int >();
  }
};

template<>
struct marshaling_traits< unsigned long int > {
  using model_type = unsigned long int;

  static variant
  marshal(unsigned long int model, marshaling_context*)
  {
    return model;
  }

  static unsigned long int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< unsigned long int >();
  }
};

// unsigned long long int

template<>
struct marshaling_traits< unsigned long long int > {
  using model_type = unsigned long long int;

  static variant
  marshal(unsigned long long int model, marshaling_context*)
  {
    return model;
  }

  static unsigned long long int
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< unsigned long long int >();
  }
};

template<>
struct marshaling_traits< float > {
  using model_type = float;

  static variant
  marshal(float model, marshaling_context*)
  {
    return model;
  }

  static float
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< float >();
  }
};

template<>
struct marshaling_traits< double > {
  using model_type = double;

  static variant
  marshal(double model, marshaling_context*)
  {
    return model;
  }

  static double
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< double >();
  }
};

template<>
struct marshaling_traits< long double > {
  using model_type = long double;

  static variant
  marshal(long double model, marshaling_context*)
  {
    return model;
  }

  static long double
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_number())
      throw std::runtime_error{ "not a number" };

    return value.get_number< long double >();
  }
};

template<>
struct marshaling_traits< std::string > {
  using model_type = std::string;

  static variant
  marshal(std::string const& model, marshaling_context*)
  {
    return model;
  }

  static std::string
  unmarshal(variant const& value, marshaling_context*)
  {
    if (!value.is_string())
      throw std::runtime_error{ "not a string" };

    return value.get_string();
  }
};

template< typename T >
struct marshaling_traits< std::vector< T > > {
  using model_type = std::vector< T >;

  static variant
  marshal(std::vector< T > const& model, marshaling_context* context)
  {
    std::vector< variant > a;

    for (auto const& j : model)
      a.emplace_back(marshaling_traits< T >::marshal(j, context));

    return a;
  }

  static std::vector< T >
  unmarshal(variant const& value, marshaling_context* context)
  {
    if (!value.is_array())
      throw std::runtime_error{ "not an array" };

    std::vector< T > model;

    for (auto const& j : value)
      model.emplace_back(marshaling_traits< T >::unmarshal(j, context));

    return model;
  }
};

template< typename T >
struct marshaling_traits< std::list< T > > {
  using model_type = std::list< T >;

  static variant
  marshal(std::list< T > const& model, marshaling_context* context)
  {
    std::vector< variant > a;

    for (auto const& j : model)
      a.emplace_back(marshaling_traits< T >::marshal(j, context));

    return a;
  }

  static std::list< T >
  unmarshal(variant const& value, marshaling_context* context)
  {
    if (!value.is_array())
      throw std::runtime_error{ "not an array" };

    std::list< T > model;

    for (auto const& j : value)
      model.emplace_back(marshaling_traits< T >::unmarshal(j, context));

    return model;
  }
};

template< typename T >
struct marshaling_traits< std::deque< T > > {
  using model_type = std::deque< T >;

  static variant
  marshal(std::deque< T > const& model, marshaling_context* context)
  {
    std::vector< variant > a;

    for (auto const& j : model)
      a.emplace_back(marshaling_traits< T >::marshal(j, context));

    return a;
  }

  static std::deque< T >
  unmarshal(variant const& value, marshaling_context* context)
  {
    if (!value.is_array())
      throw std::runtime_error{ "not an array" };

    std::deque< T > model;

    for (auto const& j : value)
      model.emplace_back(marshaling_traits< T >::unmarshal(j, context));

    return model;
  }
};

template< typename T >
struct marshaling_traits< std::set< T > > {
  using model_type = std::set< T >;

  static variant
  marshal(std::set< T > const& model, marshaling_context* context)
  {
    std::vector< variant > a;

    for (auto const& j : model)
      a.emplace_back(marshaling_traits< T >::marshal(j, context));

    return a;
  }

  static std::set< T >
  unmarshal(variant const& value, marshaling_context* context)
  {
    if (!value.is_array())
      throw std::runtime_error{ "not an array" };

    std::set< T > model;

    for (auto const& j : value)
      model.emplace(marshaling_traits< T >::unmarshal(j, context));

    return model;
  }
};

template<>
struct marshaling_traits< std::chrono::system_clock::time_point > {
  using model_type = std::string;

  static variant
  marshal(std::chrono::system_clock::time_point const& model,
          marshaling_context*)
  {
    static constexpr const char time_format[] = "%a, %d %b %Y %H:%M:%S GMT";

    std::time_t now_c = std::chrono::system_clock::to_time_t(model);

    struct std::tm tm_buf;

    std::stringstream str;
    str.imbue(std::locale{});

#ifdef _MSC_VER
    ::gmtime_s(&tm_buf, &now_c); // Stupid Microsoft.
    str << std::put_time(&tm_buf, time_format);
#else
    ::gmtime_r(&now_c, &tm_buf);
    str << std::put_time(&tm_buf, time_format);
#endif

    return str.str();
  }

  static std::chrono::system_clock::time_point
  unmarshal(variant const& value, marshaling_context*)
  {
    static constexpr const char time_format[] = "%a, %d %b %Y %H:%M:%S GMT";

    if (!value.is_string())
      throw std::runtime_error{ "not a string" };

    std::tm tm{};

    std::istringstream str{ value.get_string() };
    str.imbue(std::locale{});

    str >> std::get_time(&tm, time_format);

    if (str.fail())
      return {};

#ifdef _MSC_VER
    auto localtime = _mkgmtime(&tm);
    return std::chrono::system_clock::from_time_t(localtime);
#else
    auto localtime = timegm(&tm);
    return std::chrono::system_clock::from_time_t(localtime);
#endif
  }
};

template<typename E,
         std::string(&to_string)(E const&),
         E(&from_string)(std::string const&)>
struct enum_mapping
{
  static
  variant
  marshal(E const& value, marshaling_context*)
  {
    return to_string(value);
  }

  static
  E
  unmarshal(variant const& v, marshaling_context*)
  {
    return from_string(v.get_string());
  }
};

} // namespace code::json::marshaling

#endif
